#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "constants.h"

void create_electron(int j, int ij, double temp) {
  int Pxyz, nv;
  double sign, r = ranmc();
  double kT = kb*temp;

  if (ij != 0) sign = ij/abs(ij);
  assert ((j >= 1) && (j <= nelec));

  if (SIGEX > 0.) {
	if (r < 1./(1.+2.*exp(-DESIGEX/(kT)))) {  // D2 lower valleys
	  iv[j] = 3 + ifloor(2.*ranmc());  // --> {3,4}
	} else {                                    // D4 higher valleys
	  nv = 1 + ifloor(4.*ranmc());
	  if (nv > 2) nv += 2;
	  iv[j] = nv;                           // --> {1,2,5,6}
	}
  } else {
	iv[j] = 1 + ifloor(6.*r);               // --> pick evenly {1-6}
  }
  Pxyz = PX-1+(int)(iv[j]+1)/2;

  //if (ij != 0) fprintf(stderr, "ij = %i!\n", ij);

  do {
	//printf("iv1 = %i, j = %i\n", iv[j], j);
	Elec[PX][j] = gasdev()*sqrt(kT*mx[iv[j]]);
	//printf("iv2 = %i, j = %i\n", iv[j], j);
	if (abs(ij) == 1) {           /* X-hemi-maxwellian */
	  Elec[PX][j] = sign*fabs(Elec[PX][j]);
	} else if (abs(ij) == 2) {    /* X-FLUX */
	  Elec[PX][j] = sign*sqrt(-2.*mx[iv[j]]*kT*log(ranmc()));
	}
	Elec[PY][j] = gasdev()*sqrt(kT*my[iv[j]]);
	if (abs(ij) == 4) {           /* Y-hemi-maxwellian */
	  Elec[PY][j] = sign*fabs(Elec[PY][j]);
	} else if (abs(ij) == 8) {    /* Y-FLUX */
	  Elec[PY][j] = sign*sqrt(-2.*my[iv[j]]*kT*log(ranmc()));
	}
	Elec[PZ][j] = gasdev()*sqrt(kT*mz[iv[j]]);

	/***********************************************************************
	  The electron momentum is written in the reference frame of the CENTER
      the Brillouin zone throughout this code, so add the valley offsets.
	************************************************************************/
	Elec[PX][j] += pvx[iv[j]];
	Elec[PY][j] += pvy[iv[j]];
	Elec[PZ][j] += pvz[iv[j]];
  } while ((fabs(Elec[Pxyz][j]) > PQMAX) && (BZCLIP));

  Elec[EE][j] = energy(j);
  Elec[VX][j] = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j]);
  if (DIM == 2) {
	Elec[VY][j] = velocity(my[iv[j]],Elec[PY][j]-pvy[iv[j]],Elec[EE][j]);
  }
}

void read_input_distro(char infile[]) {
  int i = 1, j, jdup;
  FILE *in;
  char line[MAXLINE], word[MAXWORD];

  setlinebuf(stdout);
  /* open the raw data file if possible */
  in = fopen(infile,"r");
  if (in == NULL) {
	printf("\n*** ERROR:  could not read input file (%s) ***\n", infile);
	exit(-1);
  } else {
	printf("\nReading saved particle distribution (%s)\n", infile);
	printf("[make sure it's reading the correct-dimensionality file!]\n");
  }

  while (fgets(line, MAXLINE-1, in) != NULL) {
	if (nocomment(line,MAXLINE)) {
	  if (i > nelec) {
		printf("\n*** WARNING:  found too many electrons (>%i) ***\n", nelec);
		printf("*** continuing run with the first %i read ***\n\n", nelec);
		break;
	  }
	  if (DIM) {
		Elec[X][i] = atof(strtok(line, "\n\t "));
		if (DIM == 2) Elec[Y][i] = atof(strtok(NULL, "\n\t "));
		Elec[PX][i] = atof(strtok(NULL, "\t "));
	  } else {
		Elec[PX][i] = atof(strtok(line, "\t "));
	  }
	  Elec[PY][i] = atof(strtok(NULL, "\t "));
	  //	  fprintf(stderr, "got here3!\n");
	  Elec[PZ][i] = atof(strtok(NULL, "\t "));
	  //	  fprintf(stderr, "got here4!\n");
	  Elec[EE][i] = atof(strtok(NULL, "\t "));
	  //	  fprintf(stderr, "got here5!\n");
	  strlcpy(word, strtok(NULL, "\n\t "), MAXWORD);
	  if (strlen(word) == 1) iv[i] = atoi(word);
	  else iv[i] = 1 + ifloor(6.*ranmc());
	  i++;
	}
  }
  if (nelec > i-1) {
	printf("\n*** WARNING:  found fewer electrons (%i) than expected (%i) ***\n", i-1, nelec);
	printf("*** duplicating pre-existing electron info to make up the rest ***\n\n");
	for (j = i; j <= nelec; j++) {
	  jdup = (j % (i-1)) + 1;
	  if (DIM) {
		Elec[X][j] = Elec[X][jdup];
		if (DIM == 2) Elec[Y][j] = Elec[Y][jdup];
	  }
	  //	  fprintf(stderr, "got here j=%i -> %i\n",j,jdup);
	  Elec[PX][j] = Elec[PX][jdup];
	  Elec[PY][j] = Elec[PY][jdup];
	  Elec[PZ][j] = Elec[PZ][jdup];
	  Elec[EE][j] = Elec[EE][jdup];
	  iv[j] = iv[jdup];
	}
  }
}

void write_output_distro(char outfile[]) {
  int i;
  FILE *out;

  setlinebuf(stdout);
  out = fopen(outfile,"w");
  if (out == NULL) {
	fprintf(stderr, "\n*** ERROR:  could not open output file (%s) for writing ***\n", outfile);
	exit(-1);
  } else {
	fprintf(stderr, "\nWriting particle distribution to file (%s)\n", outfile);
  }
  fprintf(out, "%% tsim = %g;\n", TTOT);

  if (DIM == 1) {
	for (i = 1; i <= nelec; i++)
	  fprintf(out,"% .5e % .5e % .5e % .5e % .5f %i\n",
			  //Elec[X][i],Elec[PX][i]-pvx[iv[i]],Elec[PY][i]-pvy[iv[i]],Elec[PZ][i]-pvz[iv[i]],Elec[EE][i],iv[i]);
			  Elec[X][i],Elec[PX][i],Elec[PY][i],Elec[PZ][i],Elec[EE][i],iv[i]);
  } else if (DIM == 2) {
	for (i = 1; i <= nelec; i++)
	  fprintf(out,"% .5e % .5e % .5e % .5e % .5e % .5f %i\n",
			  //Elec[X][i],Elec[Y][i],Elec[PX][i]-pvx[iv[i]],Elec[PY][i]-pvy[iv[i]],Elec[PZ][i]-pvz[iv[i]],Elec[EE][i],iv[i]);
			  Elec[X][i],Elec[Y][i],Elec[PX][i],Elec[PY][i],Elec[PZ][i],Elec[EE][i],iv[i]);
  } else {
	for (i = 1; i <= nelec; i++)
	  fprintf(out,"% .5e  % .5e  % .5e  % .5f  %i\n",
			  Elec[PX][i],Elec[PY][i],Elec[PZ][i],Elec[EE][i],iv[i]);
  }

  fclose(out);
}

void initial_distro() {
  int i, j, k, dgrid, ix, iy;
  double temp, cumcharge=0.0, totcharge = 0.0, r, dx, dy;
  FILE *fp;
  char line[MAXLINE], *st;

  setlinebuf(stdout);

  if (DIM) {
	/******************* read grid & e-field from a file ******************/
	fp = fopen(FIELDFILE,"r");
	if (fp == NULL) {
	  printf("\n*** ERROR:  could not open grid file (%s) for reading ***\n", FIELDFILE);
	  exit(-1);
	} else printf("\nReading input grid file (%s)... ", FIELDFILE);

	/**********************************************************************
	 **************************** ONE dimension ***************************
	 *********************************************************************/
	if (DIM == 1) {
	  i = 1;
	  while (fgets(line, MAXLINE-1, fp) != NULL) {
		if (nocomment(line,MAXLINE)) {
		  if (i > BIN1) nrerror("\n*** ERROR: Not enough storage for the grid     ***\n*** Increase BIN1 in constants.h and recompile ***");
		  Xgrid[i] = atof(strtok(line, "\n\t "));
		  EFXgrid[i] = atof(strtok(NULL, "\n\t "));
		  CHGgrid[i] = atof(strtok(NULL, "\n\t "));
		  POTgrid[i] = atof(strtok(NULL, "\n\t "));
		  DOPgrid[i] = atof(strtok(NULL, "\n\t "));
		  st = strtok(NULL, "\n\t ");
		  if (st != NULL) {
			TXgrid[i] = atof(st);
		  } else {
			TXgrid[i] = T;
		  }
		  if (fabs(TXgrid[i]) > TMAX) TMAX = TXgrid[i];
		  if (fabs(DOPgrid[i]) > DOPING) DOPING = fabs(DOPgrid[i]);
		  if (fabs(DOPgrid[i]) < minDOPING) minDOPING = fabs(DOPgrid[i]);
		  i++;
		}
	  }
	  nxg = i-1;
	  XMIN = Xgrid[1];       // assume x-grid is already ordered
	  XMAX = Xgrid[nxg];
	  // these are the contacts, midway thru the first and last grid cell
	  XLCONT = 0.5*(Xgrid[1]    + Xgrid[2]);
	  XRCONT = 0.5*(Xgrid[nxg-1]+ Xgrid[nxg]);
	  // each contact has "area" DAgrid[1] and DAgrid[nxg]
	  printf("found %i X grid points.\n", nxg);
	  printf("MIN DOPING = %g ... MAX DOPING = %g\n", minDOPING, DOPING);
	  if (POISSON) {
		printf("Plasma Frequency = %g 1/sec --> ", sqrt(DOPING/epssi/mt));
		printf("min DT = %g sec\n", 1/sqrt(DOPING/epssi/mt));
		if (DT > 1/sqrt(DOPING/epssi/mt)) {
		  printf("\n*** WARNING: Time Step > 1/Plasma Frequency ***\n");
		  printf("*** You should make it less than %.2e sec ***\n\n",
				 1.0/sqrt(DOPING/epssi/mt));
		}
	  }

	  /*******************************************************************
	   ************************* TWO dimensions **************************
	   ******************************************************************/
	} else {
	  i = 1; temp = 0.;
	  while (fgets(line, MAXLINE-1, fp) != NULL) {
		if (nocomment(line,MAXLINE)) {
		  if (i > BIN1) nrerror("\n*** ERROR: Not enough storage for the grid     ***\n*** Increase BIN1 in constants.h and recompile ***");
		  Xgrid[i] = atof(strtok(line, "\n\t "));
		  if ((i > 1) && (Xgrid[i] != temp)) break;
		  temp = Xgrid[i];
		  i++;
		}
	  }
	  nyg = i-1;           /* establish nyg */
	  rewind(fp);
	  i = 1;
	  while (fgets(line, MAXLINE-1, fp) != NULL) {
		if (nocomment(line,MAXLINE)) {
		  if (i > BIN2) nrerror("\n*** ERROR: Not enough storage for the grid     ***\n*** Increase BIN2 in constants.h and recompile ***");
		  /* don't forget 1e-4 to convert microns --> cm */
		  temp = atof(strtok(line, "\n\t "));
		  if (((i-1) % nyg) == 0) Xgrid[(i-1)/nyg+1] = temp*1e-4;
		  temp = atof(strtok(NULL, "\n\t "));
		  if (i <= nyg) Ygrid[i] = temp*1e-4;
		  /* later convert two-index (i,j) to one-index (k)=(i-1)*nyg+j */
		  EFXgrid[i] = atof(strtok(NULL, "\n\t "));
		  EFYgrid[i] = atof(strtok(NULL, "\n\t "));
		  if (!((Elec[Y][i]<CHTOP) && (Elec[X][i]>OXLEF) && (Elec[X][i]<OXRIG))) {
			CHGgrid[i] = 1.;  //atof(strtok(NULL, "\n\t "));
			POTgrid[i] = 1.;  //atof(strtok(NULL, "\n\t "));
			DOPgrid[i] = 1.;  //atof(strtok(NULL, "\n\t "));
			if (fabs(DOPgrid[i]) > DOPING) DOPING = fabs(DOPgrid[i]);
		  }
		  i++;
		}
	  }
	  nxg = (i-1)/nyg;
	  XMIN = Xgrid[1];  XMAX = Xgrid[nxg];
	  YMIN = Ygrid[1];  YMAX = Ygrid[nyg];
	  printf("found %i X and %i Y grid points\n", nxg, nyg);
	  printf("MAX DOPING = %g\n", DOPING);
	}

	fclose(fp);

	printf("Xmin = %.3e (cm) .. Xmax = %.3e (cm)\n", XMIN, XMAX);
	//printvector(Xgrid,1,nxg,"xgrid.txt");
	if (DIM == 2) {
	  printf("Ymin = %.3e (cm) .. Ymax = %.3e (cm)\n", YMIN, YMAX);
	  //printvector(Ygrid,1,nyg,"ygrid.txt");
	}
	printf("[make sure the grid is really in centimeters!]\n");
	printf("Computing grid cell areas and pointer grid for later...\n");
	if (DIM == 1) {
	  for (i = 1; i <= nxg; i++) {
		// DAgrid defined to straddle the grid point, except at the edges
		if (i == 1) {
		  DAgrid[1] = .5*(Xgrid[2]-Xgrid[1]);
		} else if (i == nxg) {
		  DAgrid[nxg] = .5*(Xgrid[nxg]-Xgrid[nxg-1]);
		} else {
		  DAgrid[i] = .5*(Xgrid[i+1]-Xgrid[i-1]);
		  AA(i,i)   = epssi*(1./(Xgrid[i]-Xgrid[i-1]) + 1./(Xgrid[i+1]-Xgrid[i]));
		  AA(i,i-1) = -epssi/(Xgrid[i]  - Xgrid[i-1]);
		  AA(i,i+1) = -epssi/(Xgrid[i+1]- Xgrid[i]);
		}
		totcharge += CHGgrid[i]*DAgrid[i];
	  }
	  // now pre-calculate the upper-triangulated version of AA --> UD
	  for (i = 1; i <= 3*nxg; i++) UD[i] = AA[i];
	  for (i = 3; i <= nxg-1; i++) {
		UD(i,i) = UD(i,i) - UD(i-1,i)*UD(i,i-1)/UD(i-1,i-1);
		UD(i,i-1) = 0.0;
	  }
	} else {
	  for (i = 1; i <= nxg*nyg; i++) {
		/******** get ix and iy back from i = (ix-1)*nyg+iy *********/
		iy = i % nyg;             /* from 1 <= iy <= nyg */
		if (iy == 0) iy = nyg;
		ix = (i-iy)/nyg + 1;      /* from 1 <= ix <= nxg */
		/**************** deal w/edge-of-grid issues ****************/
		if (ix == 1) dx = .5*(Xgrid[2]-Xgrid[1]);
		else if (ix == nxg) dx = .5*(Xgrid[nxg]-Xgrid[nxg-1]);
		else dx = .5*(Xgrid[ix+1]-Xgrid[ix-1]);
		if (iy == 1) dy = .5*(Ygrid[2]-Ygrid[1]);
		else if (iy == nyg) dy = .5*(Ygrid[nyg]-Ygrid[nyg-1]);
		else dy = .5*(Ygrid[iy+1]-Ygrid[iy-1]);
		/************ compute area surrounding each cell ************/
		DAgrid[i] = dx*dy;
		totcharge += CHGgrid[i]*DAgrid[i];   /* real electrons */
	  }
	}
	/************* find out what the minimum grid spacings are *************/
	dxP = Xgrid[2]-Xgrid[1];
	for (i = 1; i <= nxg-1; i++) {
	  /****** store the grid spacings to speed up cloud_in_cell later ******/
	  DXgrid[i] = Xgrid[i+1]-Xgrid[i];
	  if (dxP > DXgrid[i]) dxP = DXgrid[i];
	}
	printf("Minimum DX = %g nm --> ", dxP*1.0e7);
	printf("max DT = %g sec\n", dxP/VLIM);
	if (DT > dxP/VLIM) {
	  printf("\n*** WARNING: Time Step too large for Xgrid spacing ***\n");
	  printf("*** Recommend you make it less than %.2e sec ***\n\n", dxP/VLIM);
	} else {
	  GLIM = 1;
	}
	if (DIM == 2) {
	  dyP = Ygrid[2]-Ygrid[1];
	  for (i = 1; i <= nyg-1; i++) {
		DYgrid[i] = Ygrid[i+1]-Ygrid[i];
		/***** store the inverse cell areas to speed up cloud_in_cell  ******/
		for (j=1;j<=nxg-1;j++) oAAgrid[(j-1)*nyg+i]= 1./(DXgrid[j]*DYgrid[i]);
		if (dyP > DYgrid[i]) dyP = DYgrid[i];
	  }
	  if (DT > dyP/VLIM) {
		printf("\n*** WARNING: Time Step too large for Ygrid spacing ***\n");
		printf("*** Recommend you make it less than %.2e sec ***\n",dyP/VLIM);
	  } else {
		GLIM = 1;
	  }
	}
	/* printvector(EFXgrid,1,nxg*nyg,"efxgrid.txt"); */
  }

  if (strlen(INPUTFILE)) {
 	read_input_distro(INPUTFILE);
  } else if (DIM) {
	if (DIM == 1)
	  printf("1D:  Assigning thermal superparticles in proportion to charge density\n\n");
	if (DIM == 2)
	  printf("2D:  Assigning thermal superparticles in proportion to charge from Medici\n\n");

	for (k = 1; k <= nelec; k++) {
#ifdef DUMBINIT
	  create_electron(k,0,T);
	  do {
		Elec[X][k] = XMIN + ranmc()*(XMAX-XMIN);
		Elec[Y][k] = YMIN + ranmc()*(YMAX-YMIN);
	  } while ((Elec[Y][k]<CHTOP) && (Elec[X][k]>OXLEF) && (Elec[X][k]<OXRIG));
#else
	  dgrid = 1;
	  cumcharge = CHGgrid[dgrid]*DAgrid[dgrid];
	  r = ranmc()*totcharge;
	  while (r > cumcharge) {     // 1 <= dgrid <= nxg (*nyg in 2D)
		dgrid++;
		cumcharge += CHGgrid[dgrid]*DAgrid[dgrid];
		if ((dgrid < 1) || (dgrid > nxg)) {
		  printf("\ndgrid = %d\n", dgrid);
		  nrerror("*** ERROR: tried to create electron outside grid (initial.c) ***\n");
		}
	  }
	  create_electron(k,0,T);
	  if (DIM == 1) {
		/* these if-else below are due to how DAgrid is defined above,
		   it straddles the grid point, except at the grid edges */
		if (dgrid == 1) {
		  Elec[X][k] = Xgrid[1]   + ranmc()*DAgrid[1];
		} else if (dgrid == nxg) {
		  Elec[X][k] = Xgrid[nxg] - ranmc()*DAgrid[nxg];
		} else {
		  Elec[X][k] = .5*(Xgrid[dgrid] + Xgrid[dgrid-1]) + ranmc()*DAgrid[dgrid];
		}
		// memorize initial positions for the NOISE calculations
#ifdef NOISE
		memX[k] = Elec[X][k];
#endif
	  } else if (DIM == 2) {
		i = 1+dgrid/nyg;          /* integer division */
		j = dgrid % nyg;
		create_electron(k,0,T);
		Elec[X][k] = Xgrid[i] + ranmc()*(Xgrid[i+1]-Xgrid[i]);
		Elec[Y][k] = Ygrid[j] + ranmc()*(Ygrid[j+1]-Ygrid[j]);
		// NOISE calculation memory to be implemented later in 2D
	  }
#endif
	}
  } else {  /* thermally initialize and uniformly assign particles in 0-D */
	printf("0D:  Starting with uniformly distributed thermal particles\n");
	for (k = 1; k <= nelec; k++) {
	  create_electron(k,0,T);
	}
  }

  if (DIM) {
	// where totcharge comes from += CHGgrid[i]*DAgrid[i] ...
	realsimscale = totcharge/((double) nelec);
	printf("One Simulated Electron = %g Real Electrons\n", realsimscale);
	printf("[assuming device dimensions not implicitly defined = 1 cm]\n");
	if (DIM == 1) {
	  NLCONT = DOPgrid[1]*DAgrid[1]/realsimscale;
	  NRCONT = DOPgrid[nxg]*DAgrid[nxg]/realsimscale;
	  printf("NLCONT = %.2f ... NRCONT = %.2f\n",NLCONT,NRCONT);
	}
	//write_output_distro("initial.txt");
	//exit(0);
  } else {
	realsimscale = (DOPING + Ni)/((double) nelec);
  }
  
  /****************** initial particle grid locations ******************/
  for (i = 1; i <= nelec; i++) {
	if (DIM) {
	  // left edge of bin
	  Gxp[i] = locate(Xgrid,1,nxg-1,Elec[X][i]);
	  // top edge of bin
	  if (DIM == 2) Gyp[i] = locate(Ygrid,1,nyg-1,Elec[Y][i]);
	}
  }

  return;
}
