#include <stdio.h>
#include "constants.h"
#include <math.h>

/*
  If RAN1 is #defined in constants.h then the random number generator
  is the one used in MOCA.  This is modeled after ran1.c from Numerical
  Recipes 2nd Ed page 280, but w/DOUBLE precision.  The function gasdev()
  generates gaussian-distributed random numbers from NR modified to use
  ranmc (with doubles) rather than ran1.

  If RANMT is #defined in constants.h then the random number generator
  from http://www.math.keio.ac.jp/matumoto/emt.html is used.

  If RAN48 is #defined in constants.h then the system's drand48() is used.

  The result is transparent to the rest of the code, which only sees and
  uses init_ranmc(long), ranmc() and gasdev().
*/

#ifdef RAN1

#define IA 16807
#define IM 2147483647
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define AM_D (1.0/IM)
#define EPS_D 3E-16
#define RNMX_D (1E0-EPS_D)

static long idum;

void init_ranmc(long j) {
  idum = j;
}

double ranmc (void) { 
  int j;
  long k;
  static long iyran=0; 
  static long riv[NTAB];
  double temp;
  if (idum <= 0 || !iyran) {     /* initialize if called with seed < 0  */
	if (-idum < 1) idum=1;    /* or if it's the first call (iyran=0) */
	else idum= -idum;
	for (j=NTAB+7;j>=0;j--) {
	  k=idum/IQ;
	  idum=IA*(idum-k*IQ)-IR*k;
	  if (idum < 0 ) idum += IM;
	  if (j < NTAB ) riv[j] = idum;
	}
	iyran=riv[0];
  }
  k=idum/IQ;
  idum=IA*(idum-k*IQ)-IR*k;
  if (idum < 0 ) idum += IM;
  j=iyran/NDIV;
  iyran=riv[j];
  riv[j]= idum;
  if ((temp=AM_D*iyran) > RNMX_D ) return RNMX_D;
  else return temp;
}

#undef IA
#undef IM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef AM_D
#undef EPS_D
#undef RNMX_D

#elif defined(RANMT)

/* 
   A C-program for MT19937, with initialization improved 2002/2/10.
   Coded by Takuji Nishimura and Makoto Matsumoto.
   This is a faster version by taking Shawn Cokus's optimization,
   Matthe Bellew's simplification, Isaku Wada's real version.

   Before using, initialize the state by using init_genrand(seed) 
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   http://www.math.keio.ac.jp/matumoto/emt.html
   email: matumoto@math.keio.ac.jp
*/

/* Period parameters */  
#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UMASK 0x80000000UL /* most significant w-r bits */
#define LMASK 0x7fffffffUL /* least significant r bits */
#define MIXBITS(u,v) ( ((u) & UMASK) | ((v) & LMASK) )
#define TWIST(u,v) ((MIXBITS(u,v) >> 1) ^ ((v)&1UL ? MATRIX_A : 0UL))

static unsigned long state[N]; /* the array for the state vector  */
static int left = 1;
static int initf = 0;
static unsigned long *next;

/* initializes state[N] with a seed */
void init_ranmc(unsigned long s) {
  int j;
  state[0]= s & 0xffffffffUL;
  for (j=1; j<N; j++) {
	state[j] = (1812433253UL * (state[j-1] ^ (state[j-1] >> 30)) + j); 
	state[j] &= 0xffffffffUL;        /* for >32 bit machines */
  }
  left = 1; initf = 1;
}

static void next_state(void) {
  unsigned long *p=state;
  int j;
  
  /* if init_ranmc() wasn't called, a default initial seed is used */
  if (initf==0) init_ranmc(5489UL);
  
  left = N;
  next = state;
  
  for (j=N-M+1; --j; p++) 
	*p = p[M] ^ TWIST(p[0], p[1]);
  
  for (j=M; --j; p++) 
	*p = p[M-N] ^ TWIST(p[0], p[1]);
  
  *p = p[M-N] ^ TWIST(p[0], state[0]);
}

/* generates a random number on (0,1)-real-interval */
double ranmc() {
  unsigned long y;

  if (--left == 0) next_state();
  y = *next++;
  
  /* Tempering */
  y ^= (y >> 11);
  y ^= (y << 7) & 0x9d2c5680UL;
  y ^= (y << 15) & 0xefc60000UL;
  y ^= (y >> 18);
  
  return ((double)y + 0.5) * (1.0/4294967296.0); 
  /* divided by 2^32 */
}

#undef N
#undef M
#undef MATRIX_A
#undef UMASK
#undef LMASK
#undef MIXBITS(u,v)
#undef TWIST(u,v)

#elif defined(RAN48)

#include <stdlib.h>

void init_ranmc(long ls) {
  return srand48(ls);
}

double ranmc() {
  return drand48();
}

#endif


double gasdev(void) {
  static int iset=0;
  static double gset;
  double fac,rsq,v1,v2;
  
#ifdef RAN1
  if (idum < 0) iset=0;
#endif
  if  (iset == 0) {
	do {
	  v1=2.0*ranmc()-1.0;
	  v2=2.0*ranmc()-1.0;
	  rsq=v1*v1+v2*v2;
	} while (rsq >= 1.0 || rsq == 0.0);
	fac=sqrt(-2.0*log(rsq)/rsq);
	gset=v1*fac;
	iset=1;
	return v2*fac;
  } else {
	iset=0;
	return gset;
  }
}
